/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Ludovic BERT & Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id: CatalinaJWebContainerServiceImpl.java,v 1.3 2003/03/07 13:02:38 danesa Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.web.catalina41;

//import java
import java.util.Vector;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.net.URL;

//import javax
import javax.naming.NamingException;

//import jonas
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.jonas.web.AbsJWebContainerServiceImpl;
import org.objectweb.jonas.web.JWebContainerServiceException;
import org.objectweb.jonas.web.lib.JarTools;
import org.objectweb.util.monolog.api.BasicLevel;

//import apache
import org.apache.commons.digester.Digester;
import org.apache.commons.digester.Rule;

//import catalina
import org.apache.catalina.Container;
import org.apache.catalina.Host;
import org.apache.catalina.core.StandardContext;
import org.apache.catalina.loader.WebappClassLoader;
import org.apache.catalina.startup.HostRuleSet;
import org.apache.catalina.startup.ContextRuleSet;

// import sax
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;


/**
 * This class provides an implementation of the Catalina service (as 
 * web container service).
 * 02/03 : Change this service for catalina version 4.1.x
 * @author Florent Benoit
 * @author Ludovic Bert
 */
public class CatalinaJWebContainerServiceImpl extends AbsJWebContainerServiceImpl {


    /**
     * Relative path of the configuration file
     */
    protected static final String CONFIG_FILE = "conf/server.xml";

    /**
     * The reference on the encapsulated embedded catalina.
     */
    protected EmbeddedCatalina embedded = null;

    /**
     * The shared extensions class loader for this server.
     */
    protected ClassLoader parentClassLoader =
        Thread.currentThread().getContextClassLoader();

    /**
     * The name of the current jonas server
     */
    private static String nameOfServer = null;


    /**
     * Initialize the Catalina service.
     * @param ctx the configuration context of the service.
     * @throws ServiceException if the initialization failed.
     */
    protected void doInit(javax.naming.Context ctx) throws ServiceException {
        super.doInit(ctx);
        boolean useSecurity = false;
        try {
            ServiceManager.getSecurityService();
            useSecurity = true;
        } catch (Exception e) {
            // nothing to do 
        }
        
        // if the user do not specify catalina.base,
        // catalina.base = catalina.home.
        String catalinaHome = System.getProperty("catalina.home");
        if (catalinaHome != null) {
            if (System.getProperty("catalina.base") == null) {
                System.setProperty("catalina.base", catalinaHome);
            }
        }
        embedded = new EmbeddedCatalina(new JCatalinaLogger(), useSecurity);

        JProp jp = null;
        try {
            jp = JProp.getInstance();
        } catch (Exception e) {
            String err = "Error when trying to get jonas name ";
            logger.log(BasicLevel.ERROR, err + " " + e.getMessage());
            throw new ServiceException(err, e);  
        }
        CatalinaJWebContainerServiceImpl.nameOfServer = jp.getValue("jonas.name", "jonas");

    }

    /**
     * Start the Catalina service in a new thread
     * @throws ServiceException if the startup failed.
     */
    public void doStart() throws ServiceException {

        // Create the digester for the parsing of the server.xml.
        Digester digester = createStartDigester();
     
        // Execute the digester for the parsing of the server.xml.
        // And configure the embedded catalina.
        File configFile = null;

        try {
            configFile = getConfigFile();
        } catch (FileNotFoundException e) {
            logger.log(BasicLevel.ERROR, "Cannot find config file " + configFile);
            throw new ServiceException(e.getMessage(), e);
        }

        try {
            InputSource is = new InputSource("file://" + configFile.getAbsolutePath());
            FileInputStream fis = new FileInputStream(configFile);
            is.setByteStream(fis);
            digester.push(embedded);
            digester.parse(is);
            fis.close();
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Error when reading config file " + configFile);
            throw new ServiceException(e.getMessage(), e);
        }


        // Finaly start embedded catalina ...
        try {
            embedded.start();
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Cannot start catalina");
            throw new ServiceException(e.getMessage(), e);
        }
        
        // ... and deploy wars of the jonas.properties
        super.doStart();
    }

    /**
     * Stop the Catalina service.
     * @throws ServiceException if the stop failed.
     */
    protected void doStop() throws ServiceException {
        // Undeploy the wars ...
        super.doStop();
        
        // ... and shut down embedded catalina 
        try {
            embedded.stop();
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Cannot stop catalina");
            throw new ServiceException(e.getMessage(), e);
        }
    }

    /**
     * Deploy a specific WAR file specified in the context.
     * @param ctx the context which contains the configuration in order 
     * to deploy the WAR.
     * @throws JWebContainerServiceException if the registration of the WAR 
     * failed.
     */
    protected void doRegisterWar(javax.naming.Context ctx) 
        throws JWebContainerServiceException {

        // Get the 4 parameters :
        //   - warURL is the URL of the war to register (required param).
        //   - contextRoot is the context root to which this application
        //     should be installed (must be unique) (required param).
        //   - hostName is the name of the host on which deploy the war
        //   - parentCL the war classloader of this war.

        URL warURL = null;
        String contextRoot = null;
        try {
            warURL = (URL) ctx.lookup("warURL");
            contextRoot = (String) ctx.lookup("contextRoot");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }

        
        ClassLoader webClassLoader = null;

        try {
            webClassLoader = (ClassLoader) ctx.lookup("parentCL");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }
        
        String hostName = null;
        try {
            hostName = (String) ctx.lookup("hostName");
        } catch (NamingException e) {
            hostName = "";
        }

        String earAppName = null;
        try {
            earAppName = (String) ctx.lookup("earAppName");
        } catch (NamingException e) {
        }

        // Install a new web application, whose web application archive is 
        // at the specified URL, into this container with the specified 
        // context root.
        // A context root of "" (the empty string) should be used for the root
        // application for this container.  Otherwise, the context root must
        // start with a slash.

        if (contextRoot.equals("/")) {
            contextRoot = "";
        }

        // install the war.
        String fileName = new File(warURL.getFile()).getName();
        String destDir = WEBAPPS_DIR + File.separator + nameOfServer + File.separator;

        // Two cases :
        // war is alone --> unpack it in : WEBAPPS_DIR/servername/filename_without_extension_dir/
        // war come from an ear --> unpack it in WEBAPPS_DIR/ear_AppName/filename_without_extension_dir/


        //Note : if the war is alone, earAppName is null.
        //ear case
        if (earAppName != null) {
            destDir += earAppName + File.separator;
        } else {
            //war alone, no need of appName in the destDir
        }

        destDir += fileName.substring(0, fileName.length() - 4);
        //Unpack the war into a directory
        JarTools.unpack(warURL.getFile(), destDir);
       
        // get the customized contexts yet created 
        Vector cvtx = embedded.getContexts();
    
        boolean deployed = false;
        for (int i = 0; i < cvtx.size(); i++) {
            StandardContext sctx = (StandardContext) cvtx.elementAt(i);
            String docBase = sctx.getDocBase();
            if (docBase.equals(contextRoot)) {
                try {
                    sctx.setParentClassLoader(webClassLoader);
                    sctx.setLoader(new JWebappLoader(webClassLoader));
                    sctx.setDocBase(destDir);
                    sctx.start();
                    logger.log(BasicLevel.DEBUG, "customized context " + contextRoot + " started");

                    //Check if the context was configured
                    if (!sctx.getConfigured()) {
                        logger.log(BasicLevel.ERROR, "context " + contextRoot + " was not configured properly. See the traces for more information.");
                    } else {
                        WebappClassLoader cl = (WebappClassLoader) sctx.getLoader().getClassLoader();
                        if (cl != null) {
                            cl.setDelegate(true);
                        }
                    }
                    deployed = true;
                } catch (Exception e) {
                    String err = "Cannot start this web application " + e;
                    logger.log(BasicLevel.ERROR, err);
                    throw new JWebContainerServiceException(err);
                }
            }
        }


        if (!deployed) {
            // no context created
            Host deployer = (Host) embedded.findHost(hostName);
            StandardContext c = (StandardContext) embedded.createContext("/" + contextRoot, destDir);
            c.setParentClassLoader(webClassLoader);
            //Define our own loader
            c.setLoader(new JWebappLoader(webClassLoader));
            deployer.addChild(c);

            //Check if the context was configured
            if (!c.getConfigured()) {
                logger.log(BasicLevel.ERROR, "context " + contextRoot + " was not configured properly. See the traces for more information.");
            } else {
                WebappClassLoader cl = (WebappClassLoader) c.getLoader().getClassLoader();
                if (cl != null) {
                    cl.setDelegate(true);
                }
                logger.log(BasicLevel.DEBUG, "context " + contextRoot + " started");
            }

            
        }

    }


    /** 
     * Undeploy a specific WAR file specified in the context.
     * @param ctx the context which contains the configuration in order
     * to undeploy a WAR.
     * @throws JWebContainerServiceException if the unregistration failed.
     */
    protected void doUnRegisterWar(javax.naming.Context ctx) 
        throws JWebContainerServiceException {

        // Get the 2 parameters :
        //   - contextRoot is the context root to be removed (required param).
        //   - hostName is the name of the host to remove the war (optional).
        String contextRoot = null;
        try {
            contextRoot = (String) ctx.lookup("contextRoot");
        } catch (NamingException e) {
            String err = "Error while getting parameter from context param ";
            logger.log(BasicLevel.ERROR, err + e.getMessage());
            throw new JWebContainerServiceException(err, e);   
        }
    
        String hostName = null;
        try {
            hostName = (String) ctx.lookup("hostName");
        } catch (NamingException e) {
            hostName = "";
        }

        // Remove the specified Context from the set of defined Contexts 
        // for its associated Host.
        Host host = embedded.findHost(hostName);
        org.apache.catalina.Context context = host.map("/" + contextRoot);
        if (context != null) {
            embedded.removeContext(context);
        } else {
            String err = "Cannot undeploy war: ";
            err = err + "Context does not exist for named path " + contextRoot;
            logger.log(BasicLevel.ERROR, err);
            throw new JWebContainerServiceException(err);
        }
    }

    /**
     * Return a File object representing the server.xml configuration file.
     * @return a File object representing the server.xml configuration file.
     * @throws FileNotFoundException if the configuration file is not found.
     */
    protected File getConfigFile() throws FileNotFoundException {
        String fileName = System.getProperty("catalina.base");
        fileName = fileName + File.separator + CONFIG_FILE;
        File file = new File(fileName);
        if (!file.exists()) {
            String err = "Can't find configuration file : " + fileName;
            throw new FileNotFoundException(err);
        }
        return (file);
    }

    /**
     * Create and configure the Digester that will be used for the xml
     * parsing of the configuration file.
     * @return Digester the digester containing the rules for the xml parsing
     * of the server.xml.
     */
    protected Digester createStartDigester() {

        // Initialize the digester
        Digester digester = new Digester();
        digester.setDebug(999);
        digester.setValidating(false);

        // Configure the actions we will be using
        //         digester.addObjectCreate("Server/GlobalNamingResources",
        //                                  "org.apache.catalina.deploy.NamingResources");
        //         digester.addSetProperties("Server/GlobalNamingResources");
        //         digester.addSetNext("Server/GlobalNamingResources",
        //                             "setGlobalNamingResources",
        //                             "org.apache.catalina.deploy.NamingResources");

        // Don't add Mbeans
        //         digester.addObjectCreate("Server/Listener",
        //                                  null, // MUST be specified in the element
        //                                  "className");
        //         digester.addSetProperties("Server/Listener");
        //         digester.addSetNext("Server/Listener",
        //                             "addLifecycleListener",
        //                             "org.apache.catalina.LifecycleListener");

        //         digester.addObjectCreate("Server/Service",
        //                                  "org.apache.catalina.core.StandardService",
        //                                  "className");
        //         digester.addSetProperties("Server/Service");
        //         digester.addSetNext("Server/Service",
        //                             "addService",
        //                             "org.apache.catalina.Service");

        digester.addObjectCreate("Server/Service/Listener",
                                 null, // MUST be specified in the element
                                 "className");
        digester.addSetProperties("Server/Service/Listener");
        digester.addSetNext("Server/Service/Listener",
                            "addLifecycleListener",
                            "org.apache.catalina.LifecycleListener");

        digester.addObjectCreate("Server/Service/Connector",
                                 "org.apache.catalina.connector.http.HttpConnector",
                                 "className");
        digester.addSetProperties("Server/Service/Connector");
        digester.addSetNext("Server/Service/Connector",
                            "addConnector",
                            "org.apache.catalina.Connector");

        digester.addObjectCreate("Server/Service/Connector/Factory",
                                 "org.apache.catalina.net.DefaultServerSocketFactory",
                                 "className");
        digester.addSetProperties("Server/Service/Connector/Factory");
        digester.addSetNext("Server/Service/Connector/Factory",
                            "setFactory",
                            "org.apache.catalina.net.ServerSocketFactory");

        digester.addObjectCreate("Server/Service/Connector/Listener",
                                 null, // MUST be specified in the element
                                 "className");
        digester.addSetProperties("Server/Service/Connector/Listener");
        digester.addSetNext("Server/Service/Connector/Listener",
                            "addLifecycleListener",
                            "org.apache.catalina.LifecycleListener");

        // Add RuleSets for nested elements
        //digester.addRuleSet(new NamingRuleSet("Server/GlobalNamingResources/"));

        digester.addRuleSet(new JEngineRuleSet("Server/Service/"));
        digester.addRuleSet(new HostRuleSet("Server/Service/Engine/"));
        digester.addRuleSet(new JContextRuleSet("Server/Service/Engine/Default", embedded));
        //digester.addRuleSet(new NamingRuleSet("Server/Service/Engine/DefaultContext/"));
        digester.addRuleSet(new JContextRuleSet("Server/Service/Engine/Host/Default", embedded));
        //digester.addRuleSet(new NamingRuleSet("Server/Service/Engine/Host/DefaultContext/"));
        digester.addRuleSet(new JContextRuleSet("Server/Service/Engine/Host/", embedded));
        //digester.addRuleSet(new NamingRuleSet("Server/Service/Engine/Host/Context/"));
        digester.addRule("Server/Service/Engine", new SetParentClassLoaderRule(digester, parentClassLoader));

        return (digester);

    }

    /**
     * Class that sets the parent class loader for the top class on the stack.
     * @author take from Catalina.
     */
    final class SetParentClassLoaderRule extends Rule {

        /**
         * The parent class loader to be set.
         */
        private ClassLoader parentClassLoader = null;


        /**
         * Construct a new action.
         * @param digester the digester need for the rule
         * @param parentClassLoader The new parent class loader
         */
        public SetParentClassLoaderRule(Digester digester,
                                        ClassLoader parentClassLoader) {
            super(digester);
            this.parentClassLoader = parentClassLoader;
        }

        /**
         * Add the requested parent class loader.
         * @param attributes the attributes.
         * @throws Exception if an error occurs.
         */
        public void begin(Attributes attributes) throws Exception {
            Container top = (Container) digester.peek();
            top.setParentClassLoader(parentClassLoader);
            
        }
    }

}





