/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: TransactionContextCORBA.jpp,v 1.5 2003/01/21 15:01:45 coqp Exp $
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_tm;

import java.io.Serializable;
import java.rmi.NoSuchObjectException;
import javax.rmi.PortableRemoteObject;
import javax.transaction.xa.Xid;
import org.omg.CORBA.Any;
import org.omg.CORBA.TCKind;
import org.omg.CosTransactions.PropagationContext;
import org.omg.CosTransactions.TransIdentity;
import org.omg.CosTransactions.otid_t;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * The propagation context is the transactional context that should be propagated
 * between programs. The structure is defined in OTS specs. 
 * In case of DAVID we rely on the std class org.omg.CosTransaction.PropagationContext
 * This class encapsulates a CORBA PropagationContext in a jonas object that
 * implements the TransactionContext interface.
 * @author Philippe Durieux
 */
public class TransactionContextCORBA implements TransactionContext {

    /**
     * @serial
     */
    private Xid xid = null;

    /**
     * embedded CORBA PropagationContext
     * @serial
     */
    private PropagationContext pctx;

    /**
     * Build a new TransactionContext when a new transaction is started.
     */
    public TransactionContextCORBA(int timeout, Coordinator coord, Xid xid) {

	TraceTm.tm.log(BasicLevel.DEBUG, "");
	// Keep xid and coord locally for efficiency
	this.xid = xid;

	// Xid is coded in the otid_t
	byte[] gtrid = xid.getGlobalTransactionId();
	byte[] bqual = xid.getBranchQualifier();
	byte[] tid = new byte[gtrid.length + bqual.length];
	System.arraycopy(bqual, 0, tid, 0, bqual.length);
	System.arraycopy(gtrid, 0, tid, bqual.length, gtrid.length);
	otid_t otid = new otid_t(xid.getFormatId(), bqual.length, tid);

	// current holds only otid_t
	TransIdentity curr = new TransIdentity(null, null, otid);

	// Create the PropagationContext
	pctx = new PropagationContext(timeout, curr, new TransIdentity[0], null);

	// Set the Coordinator.
	setCoordinator(coord);
    }

    /**
     * Build a TransactionContext from a CORBA PropagationContext
     * propagated in an incoming request in RMI-IIOP.
     */
    public TransactionContextCORBA(PropagationContext pctx) {
	this.pctx = pctx;
    }

    /**
     * Get the timeout associated with the transaction
     */
    public int getTimeout() {
	return pctx.timeout;
    }

    /**
     * Get the coordinator associated with the transaction
     */
    public Coordinator getCoordinator() {
	Any specific = pctx.implementation_specific_data;
	if (specific == null || specific.type().kind() == TCKind.tk_null) {
		return null;
	}
	return (Coordinator) PortableRemoteObject.narrow(specific.extract_Object(), Coordinator.class);
    }

    /**
     * Set the coordinator associated with the transaction
     */
    public void setCoordinator(Coordinator coord) {
	// Stub of the Coordinator is coded in the Any field.
	Any specific = org.omg.CORBA.ORB.init().create_any();
	if (coord != null) {
	    try {
		specific.insert_Object((javax.rmi.CORBA.Stub) PortableRemoteObject.toStub(coord));	
	    } catch (NoSuchObjectException e) {
		TraceTm.tm.log(BasicLevel.ERROR, "Error while coding coordinator: "+e);
	    }
	}
	pctx.implementation_specific_data = specific;
    }

    /**
     * Get the Terminator associated with the transaction
     */
    public Terminator getTerminator() {
	return (Terminator) getCoordinator();
    }

    /**
     * Get the control associated with the transaction
     */
    public Control getControl() {
	return (Control) getCoordinator();
    }

    /**
     * Get the Xid associated with the transaction
     */
    public Xid getXid() {
	if (xid == null) {
	    // Build xid from the otid_t
	    TraceTm.tm.log(BasicLevel.DEBUG, "converting otid_t to Xid");
	    otid_t otid = pctx.current.otid;
	    xid = new XidImpl(otid.formatID, otid.bqual_length, otid.tid);
	}
	return xid;
    }

    /**
     * Returns the CORBA PropagationContext (DAVID)
     */
    public PropagationContext getCorbaPropagationContext() {
	return pctx;
    }

}
